#include <stdio.h>
#include <stdlib.h>
#include "../include/utilities.h"
#include "../include/meta_explorer.h"

#define MAX_FILE_NAME_LENGTH 29
#define MAX_LINES 15
#define MAX_ROWS 3

int totalSwaps = 0;
int success = 0;

void displayResults();

void patchMetaFile(FileBlock** filesToPatch, int filesToPatchCount)
{
    printf("\nPatching Meta File...\n");
    int i = 0;

    // This is going to be used to "break" the index file
    long random_folder_num = 1;
    long random_file_num = 60556;

    // Opens the meta file
    FILE* metaFile = fopen("pad00000.meta","rb+");

    // For each file to patch
    for (i = 0; i < filesToPatchCount; i++)
    {

            // Goes to the right where the file block starts, and skips the first number (the hash)
            fseek(metaFile,filesToPatch[i]->metaOffset + sizeof(long), SEEK_SET);

            // Overrites the folder num and file num written there, to random values
            fwrite(&random_folder_num, sizeof(long),1,metaFile);
            fwrite(&random_file_num, sizeof(long),1,metaFile);
            printf("%s %s %s  (%.8ld) | %s\n", endian_convert((int)filesToPatch[i]->hash),endian_convert((int)filesToPatch[i]->folderNum),endian_convert((int)filesToPatch[i]->fileNum), filesToPatch[i]->metaOffset, filesToPatch[i]->fileName);

    }
    fclose(metaFile);
}





char* mainHeader()
{
    char* mainHeader[] = {
    "******************************************************\n",
    "*                                                    *\n",
    "*          Black Desert Online Armor Swapping        *\n",
    "*                                                    *\n",
    "******************************************************\n",
    };
    int i = 0;
    int headerLen = 0;
    for (i = 0; i < 5; i++)
    {
        headerLen += strlen(mainHeader[i]);
    }
    char* headerString = (char*)malloc(headerLen + 1);
    strcpy(headerString,mainHeader[0]);
    for (i = 1; i < 5; i++)
    {
        strcat(headerString,mainHeader[i]);
    }
    return headerString;
}

void swapArmors(char* armor1, char* armor2, FileBlock* fileBlocks, MetaFileInfo* metaFileInfo)
{
    int i = 0;
    int partsToPatchCount = 0;
    FileBlock* partsToPatch1[256];
    FileBlock* partsToPatch2[256];


    if(contains(armor1,"ub") && contains(armor2,"ub"))
    {
        for (i = 0; i < metaFileInfo->fileBlocksCount; i++)
        {
            char* currentFileClassPrefix = getClassPrefix(fileBlocks[i].fileName);
            char* currentFileTwoDigits = getTwoDigits(fileBlocks[i].fileName);
            char* currentFileFourDigits = getFourDigits(fileBlocks[i].fileName);

            char* armor1ClassPrefix = getClassPrefix(armor1);
            char* armor1TwoDigits = getTwoDigits(armor1);
            char* armor1FourDigits = getFourDigits(armor1);

            char* armor2ClassPrefix = getClassPrefix(armor2);
            char* armor2TwoDigits = getTwoDigits(armor2);
            char* armor2FourDigits = getFourDigits(armor2);

           if (fileHasExt(fileBlocks[i].fileName,"pac")
                && (contains(fileBlocks[i].folderName,"armor")
                && (strcmpi(currentFileClassPrefix,armor1ClassPrefix)) == 0)
                && (strcmpi(currentFileTwoDigits,armor1TwoDigits) == 0)
                && (strcmpi(currentFileFourDigits,armor1FourDigits) == 0))
             {
                   if (!contains(fileBlocks[i].fileName,"uw"))
                    {
                        FileBlock* partArmor1 = &fileBlocks[i];

                        char* part2Name = (char*) malloc(strlen(partArmor1->fileName));
                        strcpy(part2Name,partArmor1->fileName);

                        char* part2ClassPrefix = getClassPrefix(part2Name);
                        char* part2TwoDigits = getTwoDigits(part2Name);
                        char* part2MiddleDigits = getMiddleDigits(part2Name);
                        char* part2FourDigits = getFourDigits(part2Name);

                        strreplace(part2Name,part2ClassPrefix,armor2ClassPrefix);


                        strreplace(part2Name,part2TwoDigits,armor2TwoDigits);
                       strreplace(part2Name,part2FourDigits,armor2FourDigits);



                        if (strstr(part2Name,"_dm.pac") || strstr(part2Name,"_mul.pac")|| strstr(part2Name,"_s.pac"))
                        {

                            int j = 0;
                            int underlineCount = 0;
                            int underlineCountNeeded = 1;

                            if (strstr(part2Name,"_mul_dm.pac") ||  strstr(part2Name,"_s_dm.pac"))
                            {
                                underlineCountNeeded = 2;
                            }

                            for(j = strlen(part2Name); j >=0 ; j--)
                            {
                                if (part2Name[j] == '_')
                                {
                                    underlineCount++;

                                    if (underlineCount == underlineCountNeeded)
                                    {
                                        part2Name[j] = '.';
                                        part2Name[j+1] = 'p';
                                        part2Name[j+2] = 'a';
                                        part2Name[j+3] = 'c';
                                        part2Name[j+4] = '\0';
                                        break;
                                    }
                                }
                            }
                         }

                        FileBlock* partArmor2 = binarySearchFileBlock(part2Name,fileBlocks, metaFileInfo->fileBlocksCount);


                        if (((contains(armor2,"00_ub_0034")) && ((contains(part2Name,"sho")) || contains(part2Name,"cloak") || contains(part2Name,"lb"))) // Kibelius fix
                        || ((contains(armor2,"10_ub_0006")) && ((armor2[2] != 'm') && (armor2[2] != 'M'))))
                        {
                            // Skip

                            if (((contains(part2Name,"sho")  && contains(armor2,"_0034_01"))) ||  !contains(armor2,"_0034"))
                            {
                                partArmor2 = binarySearchFileBlock("t0072_pumpkin_ground_0001.pac",fileBlocks, metaFileInfo->fileBlocksCount);
                            }

                        }

                        if (partArmor2 == NULL)
                        {
                            if ((contains(armor2,"00_ub_0034.pac")) && (contains(partArmor1->fileName,"lb")))
                            {
                                 strreplace(part2Name,part2MiddleDigits,"sho");
                                 partArmor2 = binarySearchFileBlock(part2Name,fileBlocks, metaFileInfo->fileBlocksCount);
                            }
                            else
                            {
                                partArmor2 = binarySearchFileBlock("t0072_pumpkin_ground_0001.pac",fileBlocks, metaFileInfo->fileBlocksCount);
                            }

                        }


                        if (partArmor2 != NULL)
                        {
                            partsToPatch1[partsToPatchCount] = partArmor1;
                            partsToPatch2[partsToPatchCount] = partArmor2;

                            //printf("\n--- %s -> %s", partsToPatch2[partsToPatchCount]->fileName, partsToPatch1[partsToPatchCount]->fileName);
                            partsToPatchCount++;
                        }

                        free(part2ClassPrefix);
                        free(part2TwoDigits);
                        free(part2MiddleDigits);
                        free(part2FourDigits);

                        free(part2Name);

                    }
             }

            free(currentFileClassPrefix);
            free(currentFileTwoDigits);
            free(currentFileFourDigits);

            free(armor1ClassPrefix);
            free(armor1TwoDigits);
            free(armor1FourDigits);

            free(armor2ClassPrefix);
            free(armor2TwoDigits);
            free(armor2FourDigits);
        }
    }
    else
    {
        partsToPatchCount = 1;
        partsToPatch1[0] = binarySearchFileBlock(armor1,fileBlocks, metaFileInfo->fileBlocksCount);
        partsToPatch2[0] = binarySearchFileBlock(armor2,fileBlocks, metaFileInfo->fileBlocksCount);
    }

    char* root = bdoRootFolder();

    for (i = 0; i < partsToPatchCount; i++)
    {
        printf("\n%s -> %s", partsToPatch2[i]->fileName,partsToPatch1[i]->fileName);



        char* armor1FullPath = concatenate(root, partsToPatch1[i]->folderName);
        createPath(armor1FullPath);

        char* backup = armor1FullPath;
        armor1FullPath = concatenate(armor1FullPath, partsToPatch1[i]->fileName);

        free(backup);

        char* armor2FullPath = concatenate(root, partsToPatch2[i]->folderName);
        backup = armor2FullPath;

        armor2FullPath = concatenate(armor2FullPath, partsToPatch2[i]->fileName);
         free(backup);

        char* pathToExtract = NULL;
        if( DEBUG)
        {
            pathToExtract = concatenate("../",partsToPatch2[i]->folderName);
            extractFileCustomPath(partsToPatch2[i],pathToExtract,1);
            free(pathToExtract);
        }
        else
        {
            pathToExtract = concatenate("../",partsToPatch2[i]->folderName);
            extractFileCustomPath(partsToPatch2[i],concatenate("../",partsToPatch2[i]->folderName),0);
            free(pathToExtract);
        }


        charReplace(armor2FullPath,'/','\\');


        if (fileExists(armor1FullPath))
        {
            remove(armor1FullPath);
            if (fileExists(armor1FullPath))
            {
                printf("\nFailed to delete %s", armor1FullPath);
                PAUSE();
            }

        }

        if (rename(armor2FullPath,armor1FullPath) == 0)
        {
            if (DEBUG)
            {
                printf("\nSuccessfully renamed:\n\"%s\" \nto \n\"%s\"\n", armor2FullPath, armor1FullPath);
            }
        }
        else
        {
             printf("\n\nFailed to rename:\n\"%s\" \nto \n\"%s\"\n", armor2FullPath, armor1FullPath);
             printf("--------------------------------------------------");
             errors++;
        }
    }
    free(root);
    printf("\n\n");
    patchMetaFile(partsToPatch1, partsToPatchCount);
}

int main()
{
    char* ARMOR_SWAPPING_TXT_FILE = "patcher_resources\\armors_to_swap.txt";

    /*createPath("patcher_resources\\");

    FILE* armor_to_swap_file = NULL;
    if (!fileExists(ARMOR_SWAPPING_TXT_FILE))
    {
        armor_to_swap_file = openFile(ARMOR_SWAPPING_TXT_FILE,"wb+");
        char* defaultEntry = "THIS.pac->TURNS_TO_THIS.pac";
        fwrite(defaultEntry,strlen(defaultEntry),1,armor_to_swap_file);
        rewind(armor_to_swap_file);
    }
    else
    {
        armor_to_swap_file = openFile(ARMOR_SWAPPING_TXT_FILE,"rb+");
    }

    system(ARMOR_SWAPPING_TXT_FILE);*/
    system(ARMOR_SWAPPING_TXT_FILE);
    MetaFileInfo *metaFileInfo = getMetaFileInfo(getLatestBackup());
    FileBlock* fileBlocks = fillFileBlocks(metaFileInfo);

    if (!backupExists())
    {
        createBackup();
    }

    FILE* armor_to_swap_file = openFile(ARMOR_SWAPPING_TXT_FILE,"rb");
    createPath(concatenate(bdoRootFolder(),"object\\texture\\"));
    copyFileTo("blank_dxt1.dds","common_decals_road_04_dec.dds","patcher_resources\\texture\\",concatenate(bdoRootFolder(),"object\\texture\\"),1);
    FileBlock* blankTextureFileBlock = binarySearchFileBlock("common_decals_road_04_dec.dds",fileBlocks, metaFileInfo->fileBlocksCount);
    patchMetaFile(&blankTextureFileBlock,1);

    char armor1[128] = "";
    char armor2[128] = "";
    char c;
    int i = 0;
    int arrowFound = 0;
    int readingArmorNumber = 1;
    int result = 0;


    while (1)
    {
        result = fread(&c,1,1,armor_to_swap_file);
        // Skip Comments
        while (c == '#')
        {
            while(c != '\n')
            {
                fread(&c,1,1,armor_to_swap_file);
            }
            fread(&c,1,1,armor_to_swap_file);
        }

        if (readingArmorNumber == 1 && c != ' ' && c != '\n')
        {
            armor1[i] = c;
            i++;
        }
        else if (readingArmorNumber == 2 && c != ' ' && c != '\n')
        {
            armor2[i] = c;
            i++;
        }

        if (c == '\n' && arrowFound == 0)
        {
            //printf("\nSymbol \"->\" not found");
            i = 0;
        }
        if (c == '-')
        {
            fread(&c,1,1,armor_to_swap_file);
            if (c == '>')
            {
                armor1[i-1] = '\0';
                arrowFound = 1;
                readingArmorNumber = 2;

            }
            i = 0;
        }
        if ((c == '\n' && arrowFound == 1) || (result == 0 && arrowFound == 1))
        {
            armor2[i-1] = '\0';

            if (strcmpi(armor1,"ARMOR_TO_BE_REPLACED.PAC") != 0 && strcmpi(armor2,"ARMOR_TO_REPLACE_WITH.PAC") != 0)
            {
                printf("---------------------------------------------");
                printf("\narmor1: %s", armor1);
                printf("\narmor2: %s\n\n", armor2);

                if (!binarySearchFileBlock(armor1,fileBlocks, metaFileInfo->fileBlocksCount))
                {
                    printf("\nError: \"%s\" was not found in the reference file.\n\n", armor1);
                    errors++;
                    system("PAUSE");
                }
                else if (!binarySearchFileBlock(armor2,fileBlocks, metaFileInfo->fileBlocksCount))
                {
                    printf("\nError: \"%s\" was not found in the reference file.\n\n", armor2);
                    errors++;
                    system("PAUSE");
                }
                else
                {


                    swapArmors(armor1,armor2,fileBlocks,metaFileInfo);

                    if(errors != 0)
                    {
                        printf("\nSome errors have occurred...\nPress any key to run again with Debug Info ON for more detailed information.\n\n");
                        system("PAUSE");
                        DEBUG = 1;
                        system("cls");
                        swapArmors(armor1,armor2,fileBlocks,metaFileInfo);
                        DEBUG = 0;
                    }
                    else
                    {
                        success++;
                    }
                }


            }

            arrowFound = 0;
            readingArmorNumber = 1;
            i = 0;
            armor1[0] = '\0';
            armor2[0] = '\0';
            errors = 0;
            totalSwaps++;

        }
        if (result == 0)
        {
            break;
        }
    }
    if (totalSwaps == 0)
    {
        printf("\nNo swaps detected in the armors_to_swap.txt file.");
    }
    displayResults();
    printf("\n\n");
    system("PAUSE");
    return 0;
}



void displayResults()
{
    printColor("\n\n-------------- Overall Results --------------\n", YELLOW);
    printf("Armors Swapped: \033[32;1m%d\033[0m Not Swapped: \033[31;1m%d\033[0m",success, (int)(totalSwaps - success));

    if (totalSwaps == success)
    {
        printColor("\n\n\tAll armors have been swapped successfully!", GREEN);
    }
}

