#include "../include/meta_explorer.h"

MetaFileInfo* getMetaFileInfo(char* metaFileName)
{
    MetaFileInfo* metaFileInfo = NULL;
    printf("\nFILE       : %s",metaFileName);

    FILE* metaFile = openFile(metaFileName,"rb+");
    metaFileInfo = (MetaFileInfo*) calloc(1,sizeof(MetaFileInfo));

    fseek(metaFile,sizeof(long),SEEK_SET);
    fread(&metaFileInfo->pazCount,sizeof(long),1,metaFile);
    printf("\nPAZ_COUNT  : %ld", metaFileInfo->pazCount);


    fseek(metaFile,(metaFileInfo->pazCount * (3 * sizeof(long))),SEEK_CUR);
    fread(&metaFileInfo->fileBlocksCount,sizeof(long),1,metaFile);
    printf("\nFILES_COUNT: %ld\n", metaFileInfo->fileBlocksCount);

    metaFileInfo->fileBlocksStart = ftell(metaFile);

    printf("\nFILE_BLOCKS_START: %ld \n", metaFileInfo->fileBlocksStart);

    metaFileInfo->fileBlocksEnd = metaFileInfo->fileBlocksStart + (metaFileInfo->fileBlocksCount * ONE_REGISTRY);

    printf("FILE_BLOCKS_END  : %ld \n\n", metaFileInfo->fileBlocksEnd);

    fclose(metaFile);

    return metaFileInfo;
}

FileBlock* fillFileBlocks(MetaFileInfo* metaFileInfo)
{
    /// Allocates Fileblocks
    FileBlock* fileBlocks = NULL;
    int i = 0;

    fileBlocks = (FileBlock*)calloc(metaFileInfo->fileBlocksCount + 1, sizeof(FileBlock));

    printf("\nFilling File Blocks...\n");

    // This will open you last pad00000.meta.backup file,
    // just in case your current pad00000.meta file is already modified.
    FILE* metaFile = openFile(getLatestBackup(),"rb");

    // Go to where the file blocks start
    fseek(metaFile,metaFileInfo->fileBlocksStart,SEEK_SET);

   // Fill the File Blocks
   for (i = 0; i < metaFileInfo->fileBlocksCount; i++)
   {
        // Saves the exact byte where this registry begins (VERY IMPORTANT)
        fileBlocks[i].metaOffset = ftell(metaFile);

        // Reads the next 28 bytes of the meta file and stores it
        fread(&fileBlocks[i].hash,sizeof(long),1,metaFile);
        fread(&fileBlocks[i].folderNum,sizeof(long),1,metaFile);
        fread(&fileBlocks[i].fileNum,sizeof(long),1,metaFile);
        fread(&fileBlocks[i].pazNum,sizeof(long),1,metaFile);
        fread(&fileBlocks[i].fileOffset,sizeof(long),1,metaFile);
        fread(&fileBlocks[i].zsize,sizeof(long),1,metaFile);
        fread(&fileBlocks[i].size,sizeof(long),1,metaFile);
   }

    printf("Decrypting Folder Names...\n");
    long folders_part_length = 0;
    fread(&folders_part_length, sizeof(long),1,metaFile);
    printf("Encrypted folder names total length: %ld\n\n", folders_part_length);


/// ******************* FOLDER NAMES DECRYPTION **********************


    uint8_t *ctext = (uint8_t *) malloc(folders_part_length); // alloc to meet our need.
    uint8_t *ptext = (uint8_t *) malloc(folders_part_length);

    uint8_t *ctext_backup_pointer = ctext;
    uint8_t *ptext_backup_pointer = ptext;

    long bytesRead = fread(ctext, 1, folders_part_length, metaFile);

    if (bytesRead != folders_part_length)
    {
          printf ("Error:\nTotal number of bytes read: %ld\nExpected: %ld\n", bytesRead, folders_part_length);
    }

    const uint8_t *decrypction_key = (const uint8_t *)"\x51\xF3\x0F\x11\x04\x24\x6A\x00"; // The Black Desert ICE decryption key

    ICE_KEY *ik = ice_key_create(0); // init the key
    ice_key_set(ik, decrypction_key); // set the key

    long cuts = folders_part_length/8;
    while (cuts--){
        ice_key_decrypt(ik, ctext, ptext); // key, in (encrypted text), out (decrypted text)

        ctext +=8;
        ptext +=8;
    }

    ptext -= folders_part_length; // reset the pointer back to the beginning.



/// FOLDER NAMES FILLING


     // fileNamesArray has all the decrypted file names in order as they are read, they will be assigned to the fileBlocks later
    char** folderNamesArray = (char**) malloc(metaFileInfo->fileBlocksCount * sizeof(char*));

    // Until the \0 is found, copy the letters to the folderNamesArray[i]

    int j = 0; // The current byte of the decrypted text
    int k = 0; // The current letter of the folderNamesArray[i]
    printf("Saving Folder Names...\n");

    i = 0; // Used to advance the fileNamesArray
    long folderNameLength = 0;
    for(j = (2 * sizeof(long)) /* Skips the first 2 numbers */; j < folders_part_length; j++)
    {
        // Prevents it from going out of the fileNamesArray allocated memory
        if (i + 1 > metaFileInfo->fileBlocksCount)
        {
            break;
        }

        // If it's not a \0
        if (ptext[j] != 0)
        {
            folderNameLength++;
        }
        else // If the \0 is found
        {
            folderNamesArray[i] = (char*) malloc((folderNameLength * sizeof(char)) + 1);
            j -= folderNameLength;
            for(k = 0; k < folderNameLength; k++)
            {
                folderNamesArray[i][k] = ptext[j];
                //printf("%c", folderNamesArray[i][k]);
                j++;
            }

            folderNamesArray[i][k] = '\0'; // Ends the string

            i++; // Advances the fileNameArray position
            folderNameLength = 0;
            j += (2 * sizeof(long)); /* Skips the first 2 numbers */
        }
    }
    printf("Folder Names saved: %d\n\n", i);

    // Assigns the right File Name from the fileNameArray to the right File Block, based on the File Num
    for (i = 0; i < metaFileInfo->fileBlocksCount; i++)
    {

       // Allocates FolderName memory
        fileBlocks[i].folderName = (char*) malloc((strlen(folderNamesArray[fileBlocks[i].folderNum]) * sizeof(char)) + 1);

        // Goes to the folderNum position indicated in "fileBlocks[i].folderNum" and copies the folder name to the File Block
        strcpy(fileBlocks[i].folderName,folderNamesArray[fileBlocks[i].folderNum]);

        // Ends the string
        fileBlocks[i].folderName[strlen(folderNamesArray[fileBlocks[i].folderNum])] = '\0';
    }

    free(ctext_backup_pointer);
    free(ptext_backup_pointer);



/// ******************* FILE NAMES DECRYPTION **********************

    printf("Decrypting File Names...\n");

    long filenames_part_length = 0;
    fread(&filenames_part_length, sizeof(long),1,metaFile);
    printf("Encrypted file names total length: %ld\n\n", filenames_part_length);


    ctext = (uint8_t *) malloc(filenames_part_length); // alloc to meet our need.
    ptext = (uint8_t *) malloc(filenames_part_length);

    ctext_backup_pointer = ctext;
    ptext_backup_pointer = ptext;



    // bytesRead is the real bytes filenames_part_length that program read out, but should be same as "filenames_part_length".
    bytesRead = fread(ctext, 1, filenames_part_length, metaFile);

    if (bytesRead != filenames_part_length)
    {
          printf ("Error:\nTotal number of bytes read: %ld\nExpected: %ld\n", bytesRead, filenames_part_length);
    }

    cuts = filenames_part_length/8;
    while (cuts--){
        ice_key_decrypt(ik, ctext, ptext); // key, in (encrypted text), out (decrypted text)

        ctext +=8;
        ptext +=8;
    }

    ptext -= filenames_part_length; // reset the pointer back to the begining.


/// FILE NAMES FILLING

    // fileNamesArray has all the decrypted file names in order as they are read, they will be assigned to the fileBlocks later
    char** fileNamesArray = (char**) malloc(metaFileInfo->fileBlocksCount * sizeof(char*));

    // Until the \0 is found, copy the letters to the fileName[i]

    j = 0; // The current byte of the decrypted text
    k = 0; // The current letter of the fileName[i]
    printf("Saving file names...\n");

    i = 0; // Used to advance the fileNamesArray
    long fileNameLength = 0;
    for(j = 0; j < filenames_part_length; j++)
    {

        // Prevents it from going out of the fileNamesArray allocated memory
        if (i + 1 > metaFileInfo->fileBlocksCount)
        {
            break;
        }

        // If it's not a \0
        if (ptext[j] != 0)
        {
            fileNameLength++;
        }
        else // If the \0 is found
        {
            fileNamesArray[i] = (char*) malloc((fileNameLength * sizeof(char)) + 1);
            j -= fileNameLength;
            for(k = 0; k < fileNameLength; k++)
            {
                fileNamesArray[i][k] = ptext[j];
                //printf("%c", fileNamesArray[i][k]);
                j++;
            }

            fileNamesArray[i][k] = '\0'; // Ends the string
            i++; // Advances the fileNameArray position
            fileNameLength = 0;
        }
    }

    // Assigns the right File Name from the fileNameArray to the right File Block, based on the File Num
    for (i = 0; i < metaFileInfo->fileBlocksCount; i++)
    {

       // Allocates fileName memory
        fileBlocks[i].fileName = (char*) malloc((strlen(fileNamesArray[fileBlocks[i].fileNum]) * sizeof(char)) + 1);

        // Goes to the fileNum position indicated in "fileBlocks[i].fileNum" and copies the file name to the File Block
        strcpy(fileBlocks[i].fileName,fileNamesArray[fileBlocks[i].fileNum]);

        // Ends the string
        fileBlocks[i].fileName[strlen(fileNamesArray[fileBlocks[i].fileNum])] = '\0';
    }
    printf("File Names saved: %d (%ld less than expected)\n\n", i, metaFileInfo->fileBlocksCount - i);

    printf("\nQuick sorting File Blocks...\n");
    // QUICK SORT
    qsort(fileBlocks,metaFileInfo->fileBlocksCount,sizeof(FileBlock),compare);
    printf("Done.\n\n");


    fclose(metaFile);


    /*for (i = 0; i < metaFileInfo->fileBlocksCount; i++)
    {
        printf("%ld|%ld|%ld|%ld|%s|%s\n",fileBlocks[i].hash,fileBlocks[i].folderNum,fileBlocks[i].fileNum,fileBlocks[i].pazNum,fileBlocks[i].folderName,fileBlocks[i].fileName);
    }*/



    free(ctext_backup_pointer);
    free(ptext_backup_pointer);


    for(i = 0; i < metaFileInfo->fileBlocksCount ; i++)
    {
        free(fileNamesArray[i]);
    }
    free(fileNamesArray);
    for(i = 0; i < metaFileInfo->fileBlocksCount ; i++)
    {
        free(folderNamesArray[i]);
    }
    free(folderNamesArray);



    return fileBlocks;
}



void readBlock(FileBlock* fileBlock, FILE* fp)
{
    fileBlock->metaOffset = ftell(fp);

    fread(&fileBlock->hash,sizeof(long),1,fp);
    fread(&fileBlock->folderNum,sizeof(long),1,fp);
    fread(&fileBlock->fileNum,sizeof(long),1,fp);
    fread(&fileBlock->pazNum,sizeof(long),1,fp);
    fread(&fileBlock->fileOffset,sizeof(long),1,fp);
    fread(&fileBlock->zsize,sizeof(long),1,fp);
    fread(&fileBlock->size,sizeof(long),1,fp);

   // printf("%ld|%ld|%ld|PAZ:%ld\n",fileBlock->hash,fileBlock->folderNum,fileBlock->fileNum,fileBlock->pazNum);

}

void printBlock(FileBlock* fileBlock)
{
    printf("\nhash: %ld", fileBlock->hash);
    printf(" folder: %ld", fileBlock->folderNum);
    printf(" file: %ld", fileBlock->fileNum);
    printf(" paz_num: %ld", fileBlock->pazNum);
    printf(" offset: %ld", fileBlock->metaOffset);
    printf("| %s", fileBlock->folderName);
    printf("%s", fileBlock->fileName);
}

void printMetaFileInfo(MetaFileInfo* metaFileInfo)
{
    printf("\npazCount: %ld", metaFileInfo->pazCount);
    printf("\nfileBlocksCount: %ld", metaFileInfo->fileBlocksCount);
    printf("\nfileBlocksCount: %ld", metaFileInfo->fileBlocksCount);
    printf("\nfileBlockStart: %ld", metaFileInfo->fileBlocksStart);
    printf("\nfileBlocksEnd: %ld", metaFileInfo->fileBlocksEnd);
}

long getPazCount(char* metaFileName)
{
    long pazCount = 0;
    FILE* metaFile = openFile(metaFileName,"rb");

    fseek(metaFile,sizeof(long),SEEK_SET);
    fread(&pazCount,sizeof(long),1,metaFile);

    fclose(metaFile);
    return pazCount;
}

long getFileCount(char* metaFileName)
{
    long pazCount = 0;
    FILE* metaFile = openFile(metaFileName,"rb");

    fseek(metaFile,sizeof(long),SEEK_SET);
    fread(&pazCount,sizeof(long),1,metaFile);

    fseek(metaFile,(pazCount * (3 * sizeof(long))),SEEK_CUR);
    long fileCount = 0;

    fread(&fileCount,sizeof(long),1,metaFile);

    fclose(metaFile);

    return fileCount;
}

